# Corby Schmitz (cschmitz@anl.gov)
# Deprecated in favor of GUI
import sys
import argparse
import serial

# Dictionary for storing the list of valves
# Not all valves require directives
valves = []

# Dictionary for storing the list of directives
# Only one directive per valve
directives = []

# Function to print the main menu for the top-level
def mainMenu():
    print()
    print("1. Configure valve(s)")
    print("2. Manage directives for a valve")
    print("3. Run directives")
    print("4. Read output from valve")
    print("5. Write configuration to file")
    print("0. Quit program")

# Function to print the menu for managing valves
def valveMenu():
    print()
    print("1. Print valves")
    print("2. Add valve")
    print("3. Delete valve")
    print("0. Return to main menu")

# Function to print the menu for managing directives
def directiveMenu():
    print()
    print("1. Print directives")
    print("2. Add directive")
    print("3. Delete directive")
    print("0. Return to main menu")

# Function to take and validate a numeric input value between:
# min value (integer)
# max value (integer)
# Choice will live between the two values
def getIntChoice(min, max):
    try:
        int(min)
        int(max)
        if min > max:
            raise ValueError
    except ValueError:
        print("getIntChoice requires two integer values min and max. The value of min must be less than max")
        sys.exit(1)
    choice = 0
    candidate = input("Select option [" + str(min) + "-" + str(max) + "] >")
    try:
        choice = int(candidate)
    except ValueError:
        print("Invalid choice. Please try again.")
        return -1
    if ((choice >= min) and (choice <= max)):
            return choice
    else:
        print("Invalid choice. Please try again.")
        return -1

# Function to print the collection of valves in a human readable format
def printValves():
    if not valves:
        print("There are currently no configured valves")
    else:
        print("Currently configured valves:")
        for valve in valves:
            print("Name: " + valve["name"] + ", Port: " + valve["port"])

# Function to add a new valve to the collection via user input on the CLI
def addValve():
    valveName = input("Enter the name for the valve: ")
    portName = input("Enter the COM port for the valve: ")
    addElement = {}
    addElement["name"] = valveName
    addElement["port"] = portName
    for valve in valves:
        if valve["name"] == addElement["name"] or valve["port"] == addElement["port"]:
            print("That valve name or port is currently in use")
            return
    addElement["serial"] = serial.Serial(portName, 38400)
    if not addElement["serial"].is_open:
        addElement["serial"].open()
    valves.append(addElement)
    print("Valve added successfully")
    printValves()

# Function to add a new valve to the collection via file input at program startup
def addValveFile(valveName, portName):
    addElement = {}
    addElement["name"] = valveName
    addElement["port"] = portName
    for valve in valves:
        if valve["name"] == addElement["name"] or valve["port"] == addElement["port"]:
            return False
    addElement["serial"] = serial.Serial(portName, 38400)
    if not addElement["serial"].is_open:
        addElement["serial"].open()
    valves.append(addElement)
    return True

# Function to select and delete a valve from the collection
def delValve():
    valveName = input("Enter the valve name to delete:")
    count = 0
    for valve in valves:
        if valve["name"] == valveName:
            valves[count]["serial"].close()
            del valves[count]
            print("Valve was deleted")
            printValves()
            return
        count += 1
    print("Could not find that valve")

# Function to allow user to interactively select an option from the configure valve menu (level 2)     
def configureValve():
    while True:
        choice = -1
        while (choice < 0):
            valveMenu()
            choice = getIntChoice(0, 3)
        match choice:
            case 0:
                return
            case 1:
                printValves()
            case 2:
                addValve()
            case 3:
                delValve()
        choice = -1

# Function to print the list of existing directives in human readable format
def printDirective():
    if not directives:
        print("There are currently no configured directives")
    else:
        print("Currently configured directives:")
        for directive in directives:
            print("Name: " + directive["name"] + ", Percent Open: " + str(directive["openval"]) + \
                  ", Open Time: " + str(directive["open"]) + ", Close Time: " + str(directive["close"]) + \
                  ", Repeats: " + str(directive["repeat"]) )

# Function to allow a user to add a directive interactively via the CLI
def addDirective():
    valveName = input("Enter the name of the valve this directive is for: ")
    validValve = False
    for valve in valves:
        if valve["name"] == valveName:
            validValve = True
            break
    if not validValve:
        print("The valve provided is not found")
        return
    for directive in directives:
        if directive["name"] == valveName:
            print("A directive already exists for this valve, please delete first")
            return
    openPct = input("Enter the valve open percentage (whole number [1-100]) ")
    try:
        int(openPct)
    except ValueError:
        print("Open percentage is invalid. Please provide a whole number between 1-100")
        return
    openTime = input("Enter the open time for the valve (in whole seconds): ")
    try:
        int(openTime)
    except ValueError:
        print("Open time must be a whole integer value")
        return
    closeTime = input("Enter the close time for the valve (in whole seconds): ")
    try:
        int(openTime)
    except ValueError:
        print("Close time must be a whole integer value")
        return
    repeatNum = input("Enter the number of times to repeat the cycle: ")
    try:
        int(repeatNum)
    except ValueError:
        print("Repeat cycle must be a whole integer of 0 or more")
        return
    if int(repeatNum) < 0:
        print("Repeat cycle must be a whole integer of 0 or more")
        return
    addElement = {}
    addElement["name"] = valveName
    addElement["openval"] = round(float(openPct)/100.0, 1)
    addElement["open"] = int(openTime)
    addElement["close"] = int(closeTime)
    addElement["repeat"] = int(repeatNum)
    directiveString = str(addElement["openval"]) + "," + openTime + "," + closeTime + "," + repeatNum + "\r\n"
    addElement["bytecode"] = directiveString.encode(encoding = 'ascii', errors = 'strict')
    directives.append(addElement)
    print("Directive added successfully")
    printDirective()

# Function to add a directive via file input at program startup
def addDirectiveFile(valveName, openPct, openTime, closeTime, repeatNum):
    validValve = False
    for valve in valves:
        if valve["name"] == valveName:
            validValve = True
            break
    if not validValve:
        print("no such valve")
        return False
    for directive in directives:
        if directive["name"] == valveName:
            return False
    try:
        float(openPct)
        int(openTime)
        int(closeTime)
        int(repeatNum)
    except ValueError:
        print("int failure")
        return False
    if int(repeatNum) < 0:
        print("repeat failure")
        return False
    addElement = {}
    addElement["name"] = valveName
    addElement["openval"] = float(openPct)
    addElement["open"] = int(openTime)
    addElement["close"] = int(closeTime)
    addElement["repeat"] = int(repeatNum)
    directiveString = str(addElement["openval"]) + "," + openTime + "," + closeTime + "," + repeatNum + "\r\n"
    addElement["bytecode"] = directiveString.encode(encoding = 'ascii', errors = 'strict')
    directives.append(addElement)
    return True

# Function to select and delete a directive from the collection
def delDirective():
    valveName = input("Enter the valve name to delete a directive:")
    count = 0
    for directive in directives:
        if directive["name"] == valveName:
            del directive[count]
            print("Directive was deleted")
            printDirective()
            return
        count += 1
    print("Could not find a directive for that valve")

# Function to allow a user to select a directive management option from the menu (level 2)
def manageDirective():
    while True:
        choice = -1
        while (choice < 0):
            directiveMenu()
            choice = getIntChoice(0,3)
        match choice:
            case 0:
                return
            case 1:
                printDirective()
            case 2:
                addDirective()
            case 3:
                delDirective()

# Function to initiate the execution of a directive on a specific valve in the collection
def executeDirective(valveName):
    foundDirective = False
    for directive in directives:
        if directive["name"] == valveName:
            for valve in valves:
                if valve["name"] == valveName:
                    print("Running directive")
                    print(directive["bytecode"])
                    valve["serial"].write(directive["bytecode"])
                    output = valve["serial"].readline()
                    foundDirective = True
    if not foundDirective:
        print("No such valve exists")

# Function to read the serial output from a specific valve in the collection                
def readOutput(valveName):
    for valve in valves:
        if valve["name"] == valveName:
            print("Output from " + valveName)
            bytesWaiting = valve["serial"].inWaiting()
            if (bytesWaiting != 0):
                output = valve["serial"].read(bytesWaiting)
                print(output.decode(encoding="ascii"))
                return
    print("No output for valve " + valveName)

# Allow user to write the current collection (valves and directives) to a file on the local system
# This allows for import at a later time without having to interact with the menus
def writeFile(file):
    valveOutput = "valves:\n"
    directiveOutput = "directives:\n"
    for valve in valves:
        valveOutput += valve["name"] + ";" + valve["port"] + "\n"
    for directive in directives:
        directiveOutput += directive["name"] + ";" + str(directive["openval"]) + ";" + \
                           str(directive["open"]) + ";" + str(directive["close"]) + \
                            ";" + str(directive["repeat"]) + "\n"
    try:
        f = open(file, "w")
        f.write(valveOutput)
        f.write(directiveOutput)
    except FileNotFoundError:
        print("Filename is not valid")
# Function to read and process the contents of a configuration file and enumerate the collections (valve and directive)
# This is initiated by use of the -f <filename> program option
def readFile(file):
    try:
        f = open(file, "r")
        input = f.readline()
        if input.strip() == "valves:":
            valveLine = f.readline()
            while not valveLine.strip() == "directives:":
                (valveName, portName) = valveLine.strip().split(";")
                if addValveFile(valveName, portName):
                    print("Added valve: " + valveName + " on port: " + portName)
                else:
                    print("Could not add valve: "+ valveName + " on port: " + portName)
                valveLine = f.readline()
            directiveLine = f.readline()
            while directiveLine:
                (valveName, openPct, openTime, closeTime, repeatNum) = directiveLine.strip().split(";")
                if addDirectiveFile(valveName, openPct, openTime, closeTime, repeatNum):
                    print("Added directive for " + valveName)
                else:
                    print("Could not add directive for " + valveName)
                directiveLine = f.readline()
    except:
        print("Processing input file failed")

# Function that holds the core of the program logic with selections and branching
def mainLoop():
    while True:
        choice = -1
        while (choice < 0):
            mainMenu()
            choice = getIntChoice(0, 5)
        match choice:
            case 0:
                sys.exit(0)
            case 1:
                configureValve()
            case 2:
                manageDirective()
            case 3:
                valve = input("Enter the name of the valve to activate: ")
                executeDirective(valve)
            case 4:
                valve = input("Enter the name of the valve you want to see: ")
                readOutput(valve)
            case 5:
                file = input("Enter the name of the configuration file")
                writeFile(file)

# Parser function to build the help menu and allow for positional argument selection
parser = argparse.ArgumentParser(
                    prog="valve-control-serial",
                    description="Control application for valve management",
                    epilog="Created by Corby Schmitz (cschmitz@anl.gov)"
)

parser.add_argument("-f", '--filename')     # Optional argument that bypasses menu and runs a pre-determined directive
args = parser.parse_args()
# print(args)
if args.filename == None:
    # No file specified, so we start with a blank configuration
    print("No options specified, running in interactive mode")
else:
    # Attempt to load the filename and enumerate configuration
    readFile(args.filename)
# Initiate the main loop until user exits
mainLoop()
